/*
 * osd_mail
 *
 * Copyright (C) 2001, Andrew Rice <acr31@cam.ac.uk>
 *
 * Based on code from osd_clock.c
 * Copyright (C) 2001, Jon Beckham <leftorium@leftorium.net>
 *
 * Based on code from mailcheck.c
 * Copyright 1996, 1997, 1998 Jefferson E. Noxon <jeff@planetfall.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <xosd.h>
#include <unistd.h>
#include <ctype.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

static struct option long_options[] = {
  {"font",     1, NULL, 'f'},
  {"color",    1, NULL, 'c'},
  {"delay",    1, NULL, 'd'},
  {"format",   1, NULL, 'F'},
  {"interval", 1, NULL, 'i'},
  {"shadow",   1, NULL, 's'},
  {"top",      0, NULL, 't'},
  {"bottom",   0, NULL, 'b'},
  {"offset",   1, NULL, 'o'},
  {"help",     0, NULL, 'h'},
  {NULL,       0, NULL, 0}
};

struct linked_list {
  char* mbox;
  char* short_name;
  struct linked_list *next_ptr;
};

char *Homedir;

struct linked_list*
get_mail_boxes (FILE* rcfile) 
{
  char buf[1024], *ptr, *boxname,*sboxname,*shortname;
  struct linked_list *ll = NULL;
  struct linked_list *new_item = NULL;

  /* open the rc file and read all the entries */

  while (fgets (buf,sizeof(buf),rcfile))
    {
      /* eliminate newline */
      ptr = strchr (buf, '\n');
      if (ptr)
        *ptr = 0;

      ptr = buf;
      while(*ptr && !isspace(*ptr))
	ptr++;

      /* If it's not a blank line or comment, add it to linked list */
      if (strlen(buf) && (*ptr != '#') && (*ptr != ';'))
        {
	  new_item = malloc(sizeof(struct linked_list));

	  /* allocate a large enough string and write the mbox name into it */
	  boxname = malloc((strlen(buf)+1)*sizeof(char)); 	  /* allow for null char */
	  strcpy(boxname,buf);
	  (*new_item).mbox = boxname;
	  sboxname = strrchr(boxname,'/');

	  if (sboxname == NULL)
	    {
	      /* mbox path has no / in it !! */
	      (*new_item).short_name = boxname;
	    }
	  else {
	    shortname = malloc((strlen(sboxname)+1)*sizeof(char));
	    strcpy(shortname,sboxname+sizeof(char));
	    (*new_item).short_name = shortname;
	  }

	  (*new_item).next_ptr = ll;
	  ll = new_item;
	}
    }
  return ll;
}

FILE*
open_rcfile (void)
{
  char namebuf[256];

  snprintf (namebuf, sizeof (namebuf), "%s/.mailcheckrc", Homedir);
  if (!access (namebuf, R_OK))
    return fopen (namebuf, "r");
  return NULL;
}

int
check_for_mail (char *mailpath)
{
  struct stat st;

  if (!stat (mailpath, &st))
    {
      if ((st.st_size != 0) && (st.st_mtime > st.st_atime))
	return 1;
    }
  return 0;
}

int main (int argc, char *argv[])
{
  char c;
  
  xosd *osd;
  xosd_pos pos = XOSD_bottom;
  xosd_align hpos = XOSD_left;
  char *font = "fixed";
  char *colour = "red";
  char *message = "New mail in ";
  int delay = 3; 
  int offset = 0;
  int shadow = 2;
  int interval = 1;
  FILE *rcfile;
  
  struct linked_list *linked_list_ptr;
  struct linked_list *linked_list_element;

  Homedir = getenv ("HOME");
  if (!Homedir)
    {
      fprintf (stderr, "%s: couldn't read environment variable HOME.\n",
               argv[0]);
      return 1;
    }

  while ((c = getopt_long(argc ,argv,"f:c:d:i:s:o:v:z:h",
			  long_options, NULL)) != -1)
  {
    switch(c)
    {
      case 'f':
	font = optarg;
	break;
      case 'c':
	colour = optarg;
	break;
      case 'd':
	delay = atoi(optarg);
	break;
      case 'i':
	interval = atoi(optarg);
       break;
      case 's':
	shadow = atoi(optarg);
	break;
      case 'o':
	offset = atoi(optarg);
	break;
      case 'v':
	if(strncmp(optarg,"t",1)) {
	  pos = XOSD_top;
	}
	else if(strncmp(optarg,"m",1)) {
	  hpos = XOSD_middle;
	}
	else if (strncmp(optarg,"b",1)) {
	  hpos = XOSD_bottom;
	}	
	break;
      case 'z':
	if(strncmp(optarg,"l",1)) {
	    hpos = XOSD_left;
	}
	else if(strncmp(optarg,"c",1)) {
	  hpos = XOSD_center;
	}
	else if (strncmp(optarg,"r",1)) {
	  hpos = XOSD_right;
	}
	break;
      case 'h':
	printf("USAGE: %s [-flag args]\n"
		"\t-f\tfully qualified font.  default: fixed\n"
		"\t-c\tcolour.  default: red\n"
		"\t-s\tdrop shadow offset.  default: 2\n"
		"\t-v\tvertical position (t,m or b) (default: b)\n"
		"\t-z\thoriZontal position (l,c or r) (default: l)\n"
		"\t-o\toffset value to raise or lower around panels.  def: 0\n"
		"\t-d\tDelay (time to stay on screen when after update)\n"
	        "\t\tin seconds. default: 3\n"
		"\t-h\tthis help message\n",
		argv[0]);
	return EXIT_SUCCESS;
	break;
    }
  }

  osd = xosd_create(2);

  if (!osd)
  {
    fprintf (stderr, "Error initializing osd\n");
    return EXIT_FAILURE;
  }

  xosd_set_timeout(osd,delay);
  xosd_set_font(osd,font);
  xosd_set_colour(osd,colour);
  xosd_set_shadow_offset(osd,shadow);
  xosd_set_pos(osd,pos);
  xosd_set_align(osd,hpos);
  xosd_set_vertical_offset(osd,offset);
  rcfile = open_rcfile();

  if (! rcfile) 
    {
      fprintf(stderr,"Error: couldn't open ~/.mailcheckrc\n");
      return EXIT_FAILURE;
    }
  linked_list_ptr = get_mail_boxes(rcfile);

  while (1)
  {
    /* check the last modified time on each mailbox
       if it is different to the one stored then inspect
       it for new messages. Update the modified time and change the
       reported message if necessary */
    int result=0;
    char* output = NULL;
    int output_length=0;
    int size=0;

    linked_list_element = linked_list_ptr;
    while(linked_list_element != NULL) 
      {
	if (check_for_mail((*linked_list_element).mbox)) 
	  {
	    result=1;
	    /* calculate new string size */
	    size = strlen((*linked_list_element).short_name);
	    if (output_length == 0 )
	      {
		size += strlen(message);
	      }
	    else 
	      {
		size += sizeof(char);
	      }
	    
	    /* realloc the new string */
	    output = realloc(output,(output_length+size+1)*sizeof(char));

	    /* write the new data into the string */
	    if (output_length == 0) 
	      {
		output = strcpy(output,message);
	      }
	    else 
	      {
		output = strcat(output,",");
	      }
	    output = strcat(output,(*linked_list_element).short_name);
	   
	    /* update the size */
	    output_length+= size;
	  }
	linked_list_element = (*linked_list_element).next_ptr;      
      }   

    if (result) {
      xosd_display (osd, 1, XOSD_string, output);
      free(output);
    }

    sleep(interval);
  }

  xosd_uninit (osd);
  
  return EXIT_SUCCESS;
}
  
