/*
 * osd_mail
 *
 * Copyright (C) 2001, Andrew Rice <acr31@cam.ac.uk>
 *
 * Based on code from osd_clock.c
 * Copyright (C) 2001, Jon Beckham <leftorium@leftorium.net>
 *
 * Based on code from mailcheck.c
 * Copyright 1996, 1997, 1998 Jefferson E. Noxon <jeff@planetfall.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <xosd.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

// on my system the thing doesn't work with include <string.h>
/*char *strdup(const char *s);*/

static struct option long_options[] = {
  {"font",     1, NULL, 'f'},
  {"colour",    1, NULL, 'c'},
  {"delay",    1, NULL, 'd'},
  {"format",   1, NULL, 'F'},
  {"interval", 1, NULL, 'i'},
  {"shadow",   1, NULL, 's'},
  {"top",      0, NULL, 't'},
  {"bottom",   0, NULL, 'b'},
  {"offset",   1, NULL, 'o'},
  {"help",     0, NULL, 'h'},
  {NULL,       0, NULL, 0}
};

#define MBOX	0
#define POP3	1
#define IMAP    2

struct linked_list {
  // can be POP3,MBOX,IMAP
  int boxtype;
  char* location;
  char* name;
  int ip_addr;
  char* user;
  char* password;
  struct linked_list *next_ptr;
};

char *Homedir;

unsigned int 
get_ip(char *hostname_or_ip)
{
    struct hostent *he;
    struct in_addr ia;
    if ( !inet_aton(hostname_or_ip, &ia) )
    { // try hostname lookup
	if ( !(he = gethostbyname(hostname_or_ip)) )
	{
	    fprintf(stderr, "Couldn't get IP of '%s'\n", hostname_or_ip);
	    return(0);
	}
	ia.s_addr = *((unsigned int*)(he->h_addr_list[0]));
    }
    return(ia.s_addr);
}


struct linked_list*
get_mail_boxes (FILE* rcfile) 
{
  char buf[1024];
  char *line, *ptr,*display_name;
  int parse_error;

  struct linked_list *ll = NULL;
  struct linked_list *new_item = NULL;

  /* open the rc file and read all the entries */

  while (fgets (buf,sizeof(buf),rcfile))
    {
      parse_error=0;
      /* eliminate newline */
      ptr = strchr (buf, '\n');
      if (ptr)
        *ptr = '\0';
      else
	buf[sizeof(buf)-1]= '\0';

      /* If it's not a blank line or comment, add it to linked list */
      if (strlen (buf) && (*buf != '#'))
        {
	  new_item = malloc(sizeof(struct linked_list));

	  /* the first word on the line only indicates mailbox type so
             update ptr to point to the second word */
	  ptr = strchr(buf,' ');
	  if (ptr == NULL) parse_error=1;
	  ptr += sizeof(char);


	  if (parse_error==0) 
	    {

	      line = malloc((strlen(ptr)+1)*sizeof(char));
	      if (line == NULL) 
		{
		  fprintf(stderr,"malloc failed\n");
		  exit(-1);
		}
	      
	      strcpy(line,ptr);
	      
	      if ( strncmp(buf, "pop3",4) == 0 )
		{ /* this should be a pop3 box */
		  new_item->boxtype = POP3;
		  
		  // line should now point to a string of format "<server>
		  // <username> <password>"
		  
		  
		  new_item->location = line;
		  line = strchr(line,' ');
		  if (line == NULL) 
		    parse_error=1;
		  else
		    {
		      *line = '\0';
		      line += sizeof(char);
		      new_item->ip_addr = get_ip(new_item->location);
		      new_item->user = line;
		      line = strchr(line,' ');
		      if (line == NULL) 
			parse_error=1;
		      else
			{
			  *line = '\0';
			  line += sizeof(char);
			  new_item->password = line;
			  
			  display_name = malloc((strlen(new_item->user)+1+strlen(new_item->location)+1)*sizeof(char));

			  if (display_name == NULL) 
			    {
			      fprintf(stderr,"malloc failed!\n");
			      exit(-1);
			    }
			  *display_name='\0';
			  
			  strcat(display_name,new_item->user);
			  strcat(display_name,"@");
			  strcat(display_name,new_item->location);
			  new_item->name=display_name;
			}
		    }
		  
		  if (parse_error == 0) 
		    {
		      //printf("POP3 box: %s:%s@%s (%d)\n", new_item->user,new_item->password,new_item->location,new_item->ip_addr);
		    }
		  else 
		    {
		      fprintf(stderr, "Bad pop3 entry. Use the syntax 'pop3 <server> <username> <password>'\n");
		    }
		}
	      else if ( strncmp(buf, "imap",4) == 0) 
		{
		  /* this is an imap type box */
		  new_item->boxtype = IMAP;
		  
		  new_item->location = line;
		  line = strchr(line,' ');
		  if (line == NULL) 
		    parse_error=1;
		  else
		    {
		      *line = '\0';
		      line += sizeof(char);
		      new_item->ip_addr = get_ip(new_item->location);
		      new_item->user = line;
		      line = strchr(line,' ');
		      if (line == NULL) 
			parse_error=1;
		      else
			{
			  *line = '\0';
			  line += sizeof(char);
			  new_item->password = line;
			  
			  display_name = malloc((strlen(new_item->user)+1+strlen(new_item->location)+1)*sizeof(char));
			  
			  if (display_name == NULL) 
			    {
			      fprintf(stderr,"malloc failed!\n");
			      exit(-1);
			    }
			  *display_name='\0';
			  
			  strcat(display_name,new_item->user);
			  strcat(display_name,"@");
			  strcat(display_name,new_item->location);
			  new_item->name=display_name;
			}
		    }
	      
		  if (parse_error == 0) 
		    {
		      //printf("IMAP box: %s:%s@%s (%d)\n", new_item->user,new_item->password,new_item->location,new_item->ip_addr);
		    }
		  else 
		    {
		      fprintf(stderr, "Bad imap entry. Use the syntax 'imap <server> <username> <password>'\n");
		    }
		}
	      else if ( strncmp(buf, "mbox",4) == 0) 
		{ /* this is an mbox type box */
		  new_item->boxtype = MBOX;
		  
		  new_item->location = line;
		  display_name = strrchr(line,'/');
		  
		  if (display_name == NULL)
		    {
		      /* mbox path has no / in it !! */
		      new_item->name = line;
		    }
		  else 
		    {
		      new_item->name = display_name;
		    }
		  //printf("MBOX box: %s\n",new_item->location);
		} 
	      else 
		{
		  /* line doesn't match anything */
		  fprintf(stderr, "Bad entry in config file\n");
		  parse_error=1;
		}
	    }

	  if (parse_error==0) 
	    {
	      new_item->next_ptr = ll;
	      ll = new_item;
	    }
	  else 
	    {
	      free(new_item);
	    }
	}
    }
  return ll;
}

FILE*
open_rcfile (void)
{
  char namebuf[256];

  snprintf (namebuf, sizeof (namebuf), "%s/.osd_mailplusrc", Homedir);
  if (!access (namebuf, R_OK))
    return fopen (namebuf, "r");
  return NULL;
}

char *read_line_from_socket(int sock)
{
    static char popbuf[2001];
    fd_set rfd;
    struct timeval tv;
    int data = 0;
    int rdret;
    while ( 1 )
    {
	tv.tv_sec = 10; // read timeout in sec
	tv.tv_usec = 0;
	FD_ZERO(&rfd);
	FD_SET(sock, &rfd);
	if ( select(sock+1, &rfd, NULL, NULL, &tv) && (data < 2000) )
	{ // ok, go on and read
	    rdret = read(sock, popbuf+data, 2000-data);
	    if ( rdret > 0 )
	    {
		data += rdret;
		if ( popbuf[data-1] == 0xA )
		{
		    popbuf[data] = 0; // string termination
		    return(popbuf);
		}
	    }
	}
	else
	{
	    fprintf(stderr, "Socket read error.\n");
	    close(sock);
	    return(NULL);
	}
    }
}

int write_line_to_socket(int sock, char *line)
{
    fd_set wfd;
    struct timeval tv;
    int retval = 0;
    tv.tv_sec = 10; // wait no more than # sec if ouput buffer is full
    tv.tv_usec = 0;
    FD_ZERO(&wfd);
    FD_SET(sock, &wfd);
    if ( select(sock+1, NULL, &wfd, NULL, &tv) > 0 )
    {
	retval = write(sock, line, strlen(line)) != -1;
    }
    if ( !retval )
    {
	fprintf(stderr, "Unable to write to socket!\n");
	close(sock);
	return(0); // error
    }
    return(1);
}

int 
check_POP3_mail(struct linked_list *mailbox)
{ 

// ugly code, could be written much shorter...
    int sock;
    int msgsize;
    static char obuf[1024];
    char *input;
    struct sockaddr_in tsoa;
    if ( mailbox->boxtype != POP3 ) return(0);
    //fprintf(stderr, "Checking POP3 %s@%s\n", mailbox->user, mailbox->location);
    tsoa.sin_addr.s_addr = mailbox->ip_addr;
    tsoa.sin_port = htons(110);
    tsoa.sin_family = AF_INET;
    if ( !tsoa.sin_addr.s_addr ) return(0);
    if ( (sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1 )
    {
	fprintf(stderr, "Couldn't get socket to check %s@%s!\n", mailbox->user, mailbox->location);
	return(0);
    }
    if ( connect(sock, (struct sockaddr*)&tsoa, sizeof(struct sockaddr_in)) )
    {
	fprintf(stderr, "Couldn't connect to %s!\n", mailbox->location);
	return(0);
    }
    //fprintf(stderr, "- connected on port 110\n");
    // go on...
    // check whether server is ready
    input = read_line_from_socket(sock);
    if ( !input ) return(0);
    if ( strncmp(input, "+OK",3) != 0 )
    {
	fprintf(stderr, "POP3 error on %s\nServer returned '%s'", mailbox->location, input);
	if ( write_line_to_socket(sock, "QUIT\r\n") ) close(sock);
	return(0);
    }

    // send USER <username>
    *obuf = 0;
    strcat(obuf, "USER ");
    strcat(obuf, mailbox->user);
    strcat(obuf, "\r\n");
    if ( !write_line_to_socket(sock, obuf) ) return(0);
    input = read_line_from_socket(sock);
    if ( !input ) return(0);
    if ( strncmp(input, "+OK",3) != 0 )
    {
	fprintf(stderr, "POP3 error on %s\nServer returned '%s'", mailbox->location, input);
	if ( write_line_to_socket(sock, "QUIT\r\n") ) close(sock);
	return(0);	
    }
    // send PASS <password>
    *obuf = 0;
    strcat(obuf, "PASS ");
    strcat(obuf, mailbox->password);
    strcat(obuf, "\r\n");
    if ( !write_line_to_socket(sock, obuf) ) return(0);
    input = read_line_from_socket(sock);
    if ( !input ) return(0);
    if ( strncmp(input, "+OK",3) != 0 )
    {
	fprintf(stderr, "POP3 error on %s\nServer returned '%s'\n", mailbox->location, input);
	if ( write_line_to_socket(sock, "QUIT\r\n") ) close(sock);
	return(0);
    }
    // send STAT
    *obuf = 0;
    strcat(obuf, "STAT");
    strcat(obuf, "\r\n");
    if ( !write_line_to_socket(sock, obuf) ) return(0);
    input = read_line_from_socket(sock);
    if ( !input ) return(0);
    if ( strncmp(input, "+OK",3) != 0 )
    {
	fprintf(stderr, "POP3 error on %s\nServer returned '%s'\n", mailbox->location, input);
	if ( write_line_to_socket(sock, "QUIT\r\n") ) close(sock);
	return(0);
    }
    msgsize = 0;
    sscanf(input, "%*s %*d %d", &msgsize);
    *obuf = 0;
    strcat(obuf, "QUIT");
    strcat(obuf, "\r\n");
    if ( write_line_to_socket(sock, obuf) ) close(sock);
    if ( msgsize ) return(1);
    return(0);
}

int 
check_IMAP_mail(struct linked_list *mailbox)
{ 

// ugly code, could be written much shorter...
    int sock;
    int msgsize;
    static char obuf[1024];
    char *input;
    struct sockaddr_in tsoa;
    if ( mailbox->boxtype != IMAP ) return(0);
    //fprintf(stderr, "Checking IMAP %s@%s\n", mailbox->user, mailbox->location);
    tsoa.sin_addr.s_addr = mailbox->ip_addr;
    tsoa.sin_port = htons(143);
    tsoa.sin_family = AF_INET;
    if ( !tsoa.sin_addr.s_addr ) return(0);
    if ( (sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1 )
    {
	fprintf(stderr, "Couldn't get socket to check %s@%s!\n", mailbox->user, mailbox->location);
	return(0);
    }
    if ( connect(sock, (struct sockaddr*)&tsoa, sizeof(struct sockaddr_in)) )
    {
	fprintf(stderr, "Couldn't connect to %s!\n", mailbox->location);
	return(0);
    }
    //fprintf(stderr, "- connected on port 110\n");
    // go on...
    // check whether server is ready
    input = read_line_from_socket(sock);
    if ( !input ) return(0);
    if ( strncmp(input, "* OK",4) != 0 )
    {
	fprintf(stderr, "IMAP error on %s\nServer returned '%s'", mailbox->location, input);
	if ( write_line_to_socket(sock, "A005 LOGOUT\r\n") ) close(sock);
	return(0);
    }

    // send A001 LOGIN <username> <password>
    *obuf = 0;
    strcat(obuf, "A001 LOGIN ");
    strcat(obuf, mailbox->user);
    strcat(obuf," ");
    strcat(obuf,mailbox->password);
    strcat(obuf, "\r\n");
    if ( !write_line_to_socket(sock, obuf) ) return(0);
    input = read_line_from_socket(sock);
    if ( !input ) return(0);
    if ( strncmp(input, "A001 OK",7) != 0 )
    {
	fprintf(stderr, "IMAP error on %s\nServer returned '%s'", mailbox->location, input);
	if ( write_line_to_socket(sock, "A005 LOGOUT\r\n") ) close(sock);
	return(0);	
    }
    // send A003 STATUS INBOX (UNSEEN)
    *obuf = 0;
    strcat(obuf, "A003 STATUS INBOX (UNSEEN)\r\n");
    if ( !write_line_to_socket(sock, obuf) ) return(0);
    input = read_line_from_socket(sock);
    if ( !input ) return(0);
    if ( strncmp(input, "A003 STATUS",11) == 0 )
    {
	fprintf(stderr, "IMAP error on %s\nServer returned '%s'\n", mailbox->location, input);
	if ( write_line_to_socket(sock, "A005 LOGOUT\r\n") ) close(sock);
	return(0);
    }
    msgsize = 0;
    sscanf(input, "* %*s %*s (UNSEEN %d)", &msgsize);
    if ( write_line_to_socket(sock, "A005 LOGOUT\r\n") ) close(sock);
    if ( msgsize ) return(1);
    return(0);
}

int
check_MBOX_mail (struct linked_list *mailbox)
{
  struct stat st;

  if (!stat (mailbox->location, &st))
    {
      if ((st.st_size != 0) && (st.st_mtime > st.st_atime))
	return 1;
    }
  return 0;
}


int main (int argc, char *argv[])
{
  char c;
  
  xosd *osd;
  xosd_pos pos = XOSD_bottom;
  xosd_align hpos = XOSD_left;

  char *font = "fixed";
  char *colour = "red";
  char *message = "New mail in ";
  int delay = 300; // display for 3 min
  int offset = 0;
  int shadow = 1;
  int interval = 600;
  FILE *rcfile;
  
  struct linked_list *linked_list_ptr;
  struct linked_list *linked_list_element;

  Homedir = getenv ("HOME");
  if (!Homedir)
    {
      fprintf (stderr, "%s: couldn't read environment variable HOME.\n",
               argv[0]);
      return 1;
    }
  
  while ((c = getopt_long(argc ,argv,"f:c:d:i:s:o:v:z:h",
			  long_options, NULL)) != -1)
    {
      switch(c)
	{
	case 'f':
	  font = optarg;
	  break;
	case 'c':
	  colour = optarg;
	  break;
	case 'd':
	  delay = atoi(optarg);
	  break;
	case 'i':
	  interval = atoi(optarg);
	  break;
	case 's':
	  shadow = atoi(optarg);
	  break;
	case 'o':
	  offset = atoi(optarg);
	  break;
	case 'v':
	  if(strncmp(optarg,"t",1)) {
	    pos = XOSD_top;
	  }
	  else if(strncmp(optarg,"m",1)) {
	    hpos = XOSD_middle;
	  }
	  else if (strncmp(optarg,"b",1)) {
	    hpos = XOSD_bottom;
	  }	
	  break;
	case 'z':
	  if(strncmp(optarg,"l",1)) {
	    hpos = XOSD_left;
	  }
	  else if(strncmp(optarg,"c",1)) {
	    hpos = XOSD_center;
	  }
	  else if (strncmp(optarg,"r",1)) {
	    hpos = XOSD_right;
	  }
	  break;
	case 'h':
	  printf("USAGE: %s [-flag args]\n"
		 "\t-f\tfully qualified font.  default: fixed\n"
		 "\t-c\tcolour.  default: red\n"
		 "\t-s\tdrop shadow offset.  default: 2\n"
		 "\t-v\tvertical position (t,m or b) (default: b)\n"
		 "\t-z\thoriZontal position (l,c or r) (default: l)\n"
		 "\t-o\toffset value to raise or lower around panels.  def: 0\n"
		 "\t-d\tDelay (time to stay on screen when after update)\n"
		 "\t\tin seconds. default: 3\n"
		 "\t-h\tthis help message\n",
		 argv[0]);
	  return EXIT_SUCCESS;
	  break;
	}
    }

  osd = xosd_create(2);
  
  if (!osd)
    {
      fprintf (stderr, "Error initializing osd\n");
      return EXIT_FAILURE;
  }

  xosd_set_timeout(osd,delay);
  xosd_set_font(osd,font);
  xosd_set_colour(osd,colour);
  xosd_set_shadow_offset(osd,shadow);
  xosd_set_pos(osd,pos);
  xosd_set_align(osd,hpos);
  xosd_set_vertical_offset(osd,offset);
  rcfile = open_rcfile();

  rcfile = open_rcfile();

  if (! rcfile) 
    {
      fprintf(stderr,"Error: couldn't open ~/.osd_mailplusrc\n");
      return EXIT_FAILURE;
    }
  linked_list_ptr = get_mail_boxes(rcfile);

  while (1)
  {
    /* check the last modified time on each mailbox
       if it is different to the one stored then inspect
       it for new messages. Update the modified time and change the
       reported message if necessary */
    int result=0;
    char* output = NULL;
    int output_length=0;
    int size=0;
    int ret;

    linked_list_element = linked_list_ptr;
    while(linked_list_element != NULL) 
      {
        switch ( linked_list_element->boxtype )
	  {
	  case POP3:
	    ret = check_POP3_mail(linked_list_element);
	    break;
	  case IMAP:
	    ret = check_IMAP_mail(linked_list_element);
	    break;
	  case MBOX:
	  default:
	    ret = check_MBOX_mail(linked_list_element);
	  }
	if (ret) 
	  {
	    result=1;
	    /* calculate new string size */
	    size = strlen(linked_list_element->name)+1;
	    if (output_length == 0 )
	      {
		size += strlen(message);
	      }
	    else 
	      {
		size += sizeof(char);
	      }
	    
	    /* realloc the new string */
	    output = realloc(output,(output_length+size+1)*sizeof(char));

	    /* write the new data into the string */
	    if (output_length == 0) 
	      {
		output = strcpy(output,message);
	      }
	    else 
	      {
		output = strcat(output,",");
	      }
	    
	    output = strcat(output,(*linked_list_element).name);
	  
	   
	    /* update the size */
	    output_length+= size;
	  }
	linked_list_element = (*linked_list_element).next_ptr;      
      }   

    if (result) {
      xosd_display (osd, 1, XOSD_string, output);
      free(output);
    }

    sleep(interval);
  }

  xosd_uninit (osd);
  
  return EXIT_SUCCESS;
}



